"""
Provides effortless work with YooMoney API using asynchronous requests.

"""
from __future__ import annotations

import asyncio
import warnings
from datetime import datetime
from typing import List, Dict, Any, Union, Optional, Tuple, cast

from glQiwiApi.core.abc.wrapper import Wrapper
from glQiwiApi.core.constants import NO_CACHING
from glQiwiApi.core.mixins import DataMixin, ContextInstanceMixin
from glQiwiApi.core.request_service import RequestService
from glQiwiApi.core.session.holder import AbstractSessionHolder
from glQiwiApi.types import (
    AccountInfo,
    OperationType,
    Operation,
    OperationDetails,
    PreProcessPaymentResponse,
    Payment,
    IncomingTransaction,
)
from glQiwiApi.types.yoomoney.types import Card
from glQiwiApi.utils.exceptions import CantParseUrl, InvalidPayload
from glQiwiApi.utils.payload import (
    parse_auth_link,
    retrieve_base_headers_for_yoomoney,
    parse_iterable_to_list_of_objects,
    check_params,
    parse_amount,
    format_transactions_payload,
    make_payload, filter_none,
)
from glQiwiApi.utils.validators import String
from glQiwiApi.yoo_money.settings import YooMoneyRouter, YooMoneyMethods


class YooMoneyAPI(Wrapper, DataMixin, ContextInstanceMixin["YooMoneyAPI"]):
    """
    That class implements processing requests to YooMoney
    It is convenient in that it does not just give json such objects,
    and all this converts into pydantic models.
    To work with this class, you need to register a token,
    using the guide on the official github of the project

    """
    api_access_token = String(optional=False)

    def __init__(
            self,
            api_access_token: str,
            cache_time: Union[float, int] = NO_CACHING,
            session_holder: Optional[AbstractSessionHolder[Any]] = None,
    ) -> None:
        """
        The constructor accepts a token obtained from the method class get_access_token
         and the special attribute without_context

        :param api_access_token: api token for requests
        :param cache_time: Time to cache requests in seconds,
         default 0, respectively the request will not use the cache by default
        :param session_holder: obtains session and helps to manage session lifecycle. You can pass
                               your own session holder, for example using httpx lib and use it
        """
        self.api_access_token = api_access_token
        self._router = YooMoneyRouter()
        self._request_service = RequestService(self._router.config.ERROR_CODE_MESSAGES,
                                               cache_time, session_holder=session_holder)

        self.set_current(self)

    def get_request_service(self) -> RequestService:
        return self._request_service

    def _auth_token(self, headers: Dict[Any, Any]) -> Dict[Any, Any]:
        headers["Authorization"] = headers["Authorization"].format(
            token=self.api_access_token
        )
        return headers

    @classmethod
    async def build_url_for_auth(
            cls, scope: List[str], client_id: str, redirect_uri: str = "https://example.com"
    ) -> Optional[str]:
        """
        Method to get the link for further authorization and obtaining a token

        :param scope: OAuth2 authorization of the application by the user,
         the rights are transferred by the list.
        :param client_id: application id, type string
        :param redirect_uri: a funnel where the temporary code that you need will go to
         to get the main token
        :return: the link to follow
         and make authorization via login / password
        """
        router = YooMoneyRouter()
        headers = retrieve_base_headers_for_yoomoney()
        params = {
            "client_id": client_id,
            "response_type": "code",
            "redirect_uri": redirect_uri,
            "scope": " ".join(scope),
        }
        url = router.build_url(YooMoneyMethods.BUILD_URL_FOR_AUTH)
        rs = RequestService(error_messages=router.config.ERROR_CODE_MESSAGES)
        response = await rs.text_content(url, "POST", headers=headers, data=params)
        try:
            return parse_auth_link(response)
        except IndexError:
            raise CantParseUrl(
                "Could not find the authorization link in the response from "
                "the api, check the client_id value"
            )
        finally:
            await rs.shutdown()

    @classmethod
    async def get_access_token(
            cls,
            code: str,
            client_id: str,
            redirect_uri: str = "https://example.com",
            client_secret: Optional[str] = None,
    ) -> str:
        """
        Method for obtaining a token for requests to the YooMoney API

        :param code: the temporary code that was obtained in the base_authorize method
        :param client_id: application id, type string
        :param redirect_uri: the funnel where the temporary code will go,
         which is needed to get the main token
        :param client_secret: The secret word for authenticating the application.
         Specified if the service is registered with authentication.
        :return: YooMoney API TOKEN
        """
        router = YooMoneyRouter()
        headers = retrieve_base_headers_for_yoomoney(content_json=True)
        params = {
            "code": code,
            "client_id": client_id,
            "grant_type": "authorization_code",
            "redirect_uri": redirect_uri,
            "client_secret": client_secret,
        }
        rs = RequestService(error_messages=router.config.ERROR_CODE_MESSAGES)
        try:
            response = await rs.api_request(
                "POST",
                YooMoneyMethods.GET_ACCESS_TOKEN,
                router,
                headers=headers,
                data=params,
            )
        finally:
            await rs.shutdown()
        return cast(str, response["access_token"])

    async def revoke_api_token(self) -> Optional[Dict[str, bool]]:
        """
        Method for revoking the rights of a token, while all its rights also disappear
        Documentation:
        https://yoomoney.ru/docs/wallet/using-api/authorization/revoke-access-token
        """
        headers = self._auth_token(retrieve_base_headers_for_yoomoney(auth=True))
        return await self._request_service.api_request(
            "POST", YooMoneyMethods.REVOKE_API_TOKEN, self._router, headers=headers
        )

    @property
    async def account_info(self) -> AccountInfo:
        warnings.warn(
            "`QiwiWrapper.get_account_info` property is deprecated, and will be removed in next versions, "
            "use `QiwiWrapper.retrieve_account_info(...)` instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return await self.retrieve_account_info()

    @classmethod
    def create_pay_form(cls, receiver: str,
                        quick_pay_form: str,
                        targets: str,
                        payment_type: str,
                        amount: Union[int, float],
                        form_comment: Optional[str] = None,
                        short_dest: Optional[str] = None,
                        label: Optional[str] = None,
                        comment: Optional[str] = None,
                        success_url: Optional[str] = None,
                        need_fio: Optional[bool] = None,
                        need_email: Optional[bool] = None,
                        need_phone: Optional[bool] = None,
                        need_address: Optional[bool] = None) -> str:
        """
        The YooMoney form is a set of fields with information about a transfer.
        You can embed payment form into your interface (for instance, a website or blog).
        When the sender pushes the button, the details from the form are sent to YooMoney
        and an order for a transfer to your wallet is initiated.

        Detail docs: https://yoomoney.ru/docs/payment-buttons/using-api/forms?lang=en

        Possible values for quick_pay_form:
           * shop - for a multi purpose form;
           * small - for a button;
           * donate - for a charity form.

        @param receiver: Number of the YooMoney wallet which money from senders is credited to.
        @param quick_pay_form:
        @param targets: Payment purpose.
        @param payment_type: Payment method. Possible values: PC, AC, MC
        @param amount: 	Transfer amount (the amount debited from the sender).
        @param form_comment: Name of the transfer in sender’s history
                (for transfers from a wallet or linked bank card). Displayed in sender’s wallet.
                The simplest way to create it is to combine the names of the store and product.
                For instance:  My Store: white valenki boots
        @param short_dest: Name of the transfer on the confirmation page.
               We recommend using the same name as formcomment
        @param label: The label that a site or app assigns to a certain transfer.
                      For instance, a code or order identifier may be used for this label.
        @param comment: The field in which you can send sender’s comments.
        @param success_url: URL where the user is redirected after the transfer.
        @param need_fio: Sender’s full name required.
        @param need_email: Sender’s email address required.
        @param need_phone: Sender’s phone number required.
        @param need_address: Sender’s address required.
        @return: link to payment form
        """
        payload = make_payload(**{
            "receiver": receiver,
            "quickpay-form": quick_pay_form,
            "targets": targets,
            "paymentType": payment_type,
            "sum": amount,
            "formcomment": form_comment,
            "short-dest": short_dest,
            "label": label,
            "comment": comment,
            "successURL": success_url,
            "need-fio": need_fio,
            "need-email": need_email,
            "need-phone": need_phone,
            "need-address": need_address
        })
        router = YooMoneyRouter()
        base_url = router.build_url(YooMoneyMethods.QUICK_PAY_FORM)
        params = "".join(f"&{key}={value}" for key, value in payload.items())
        return base_url + params

    async def retrieve_account_info(self) -> AccountInfo:
        """
        Method for getting information about user account
        Detailed documentation:
        https://yoomoney.ru/docs/wallet/user-account/account-info

        :return: объект AccountInfo
        """
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        response = await self._request_service.api_request(
            "POST", YooMoneyMethods.ACCOUNT_INFO, self._router, headers=headers
        )
        return AccountInfo.parse_obj(response)

    async def transactions(
            self,
            operation_types: Optional[
                Union[List[OperationType], Tuple[OperationType, ...]]
            ] = None,
            start_date: Optional[datetime] = None,
            end_date: Optional[datetime] = None,
            start_record: Optional[int] = None,
            records: int = 30,
            label: Optional[Union[str, int]] = None,
    ) -> List[Operation]:
        """
        More details:
        https://yoomoney.ru/docs/wallet/user-account/operation-history\n
        The method allows you to view the history of transactions (in whole or in part) in page mode.
        History records are displayed in reverse chronological order from most recent to earlier.

        Possible values: \n
        DEPOSITION — refill ;\n
        PAYMENT — consumption;\n
        INCOMING(incoming-transfers-unaccepted) — unaccepted incoming P2P transfers of any type.\n

        :param operation_types: Operation type
        :param label: string.
         Selection of payments by tag value.
         Selects payments that have a specified parameter value
         label of the request-payment call.
        :param start_date: Show operations from the moment in time
         (operations equal to start_date or later)
         If the parameter is absent, all operations are displayed.
        :param end_date: Output operations up to the point in time
            (operations older than end_date).
            If the parameter is absent, all operations are displayed.
        :param start_record: If the parameter is present, then operations will be displayed starting
          from start_record number.
         Operations are numbered from 0. More about paginated list output
        :param records:	The number of transaction history records requested.
         Valid values are from 1 to 100, the default is 30.
        """
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        data = {"records": records, "label": label}
        payload = format_transactions_payload(
            data, records, operation_types, start_date, end_date, start_record
        )
        response = await self._request_service.api_request(
            "POST",
            YooMoneyMethods.TRANSACTIONS,
            self._router,
            headers=headers,
            data=filter_none(payload),
        )
        return parse_iterable_to_list_of_objects(
            iterable=cast(List[Any], response.get("operations")), model=Operation
        )

    async def transaction_info(self, operation_id: str) -> OperationDetails:
        """
        Allows you to get detailed information about the operation from the history.
        Required token rights: operation-details.
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/user-account/operation-details

        :param operation_id: Operation ID
        """
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        response = await self._request_service.api_request(
            "POST",
            YooMoneyMethods.TRANSACTION_INFO,
            self._router,
            data={"operation_id": operation_id},
            headers=headers,
        )
        return OperationDetails.parse_obj(response)

    async def _pre_process_payment(
            self,
            to_account: str,
            amount: Union[int, float],
            pattern_id: str = "p2p",
            comment_for_history: Optional[str] = None,
            comment_for_receiver: Optional[str] = None,
            protect: bool = False,
            expire_period: int = 1,
    ) -> PreProcessPaymentResponse:
        """
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/process-payments/request-payment
        Creation of payment, verification of parameters and acceptance
        payment by the store or transfer of funds to the user account of YooMoney.
        This method is not recommended to be used directly, much
        it's easier to use send.
        Required token rights: to-account ("recipient id", "id type")

        :param pattern_id: Payment pattern ID
        :param to_account: string ID of the transfer recipient
         (account number, phone number or email).
        :param amount: Amount to be received
         (the invoice will be sent to the recipient's account after payment).
        :param comment_for_history: Comment to the translation,
         displayed in the sender's history.
        :param comment_for_receiver: string Comment to the translation,
         displayed to the recipient.
        :param protect: The value of the parameter is true - a sign that
         that the transfer is protected by a protection code.
         By default, there is no parameter (normal translation).
        :param expire_period: The number of days during which:
            the recipient of the transfer can enter the protection code
            and receive a transfer to your account,
            the recipient of the transfer on demand can receive the transfer.
            The parameter value must be in the range from 1 to 365.
            Optional parameter. The default is 1.
        """
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        payload = {
            "pattern_id": pattern_id,
            "to": to_account,
            "amount_due": amount,
            "comment": comment_for_history,
            "message": comment_for_receiver,
            "expire_period": expire_period,
            "codepro": protect,
        }
        response = await self._request_service.api_request(
            "POST",
            YooMoneyMethods.PRE_PROCESS_PAYMENT,
            self._router,
            headers=headers,
            data=payload,
        )
        return PreProcessPaymentResponse.parse_obj(response)

    async def send(
            self,
            to_account: str,
            amount: Union[int, float],
            money_source: str = "wallet",
            pattern_id: str = "p2p",
            cvv2_code: str = "",
            card_type: Optional[str] = None,
            protect: bool = False,
            comment_for_history: Optional[str] = None,
            comment: Optional[str] = None,
            expire_period: int = 1,
    ) -> Payment:
        """
        A method for sending money to another person's account or card.
        This function makes 2 requests at once, because of this you may feel a slight loss in performance,
          you can use the method
        _pre_process_payment and get the PreProcessPaymentResponse object,
        which contains information about a still unconfirmed payment \n
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/process-payments/process-payment

        :param pattern_id: Payment pattern ID
        :param to_account: string ID of the transfer recipient
         (account number, phone number or email).
        :param amount: Amount to be received
         (the invoice will be sent to the recipient's account after payment). MINIMUM AMOUNT 2.
        :param money_source: The requested payment method.
         wallet - from the user's account, if you want to use a card,
         then you will need to pass card_type
         to search for a card in the list of your bank cards,
         and also optionally cvv2 code for making a payment
        :param comment_for_history: Comment to the translation,
         displayed in the sender's history.
        :param card_type: Bank card type, you need to fill in,
         only if you want to debit funds from your card
        :param cvv2_code: optional, may not be passed, however
         if payment by card is required,
         the parameter should be passed
        :param comment: Comment on the transfer, displayed to the recipient.
        :param protect: The value of the parameter is true - a sign that
         that the transfer is protected by a protection code.
         By default, there is no parameter (normal translation).
        :param expire_period: Number of days during which:
            the recipient of the transfer can enter the protection code and
            receive a transfer to your account,
            the recipient of the transfer on demand can receive the transfer.
            The parameter value must be in the range from 1 to 365.
            Optional parameter. The default is 1.


        """
        if amount < 2:
            raise InvalidPayload(
                "Enter the amount that is more than the minimum (2 or more)"
            )
        pre_payment = await self._pre_process_payment(
            to_account=to_account,
            amount=amount,
            comment_for_history=comment_for_history,
            comment_for_receiver=comment,
            expire_period=expire_period,
            protect=protect,
            pattern_id=pattern_id,
        )
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        payload = {"request_id": pre_payment.request_id, "money_source": "wallet"}
        if (
                money_source == "card"
                and isinstance(pre_payment, PreProcessPaymentResponse)  # noqa: W503
                and pre_payment.money_source.cards.allowed == "true"  # type: ignore  # noqa: W503
        ):
            if not isinstance(card_type, str):
                cards = cast(Card, pre_payment.money_source.cards)  # type: ignore
                payload.update(
                    {"money_source": cards.items[0].item_id, "csc": cvv2_code}
                )
            else:
                cards = cast(Card, pre_payment.money_source.cards).items  # type: ignore
                for card in cards:
                    if card.item_type == card_type:  # type: ignore
                        payload.update(
                            {
                                "money_source": card.item_id,  # type: ignore
                                "csc": cvv2_code,
                            },
                        )
        response = await self._request_service.api_request(
            "POST",
            YooMoneyMethods.PROCESS_PAYMENT,
            self._router,
            headers=headers,
            data=payload,
        )
        return Payment.parse_obj(response).initialize(pre_payment.protection_code)

    async def get_balance(self) -> float:
        return (await self.retrieve_account_info()).balance

    async def accept_incoming_transaction(
            self, operation_id: str, protection_code: str
    ) -> IncomingTransaction:
        """
        Acceptance of incoming transfers protected by a protection code,
        if you passed the protect parameter to the send method
        Number of reception attempts
        incoming transfer with a protection code is limited.
        When the number of attempts is exhausted, the transfer is automatically rejected
        (the transfer is returned to the sender).
        More detailed documentation:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-accept

        :param operation_id: Operation identifier,
         the value of the operation_id parameter of the history () method response
        :param protection_code: Protection code. A string of 4 decimal digits.
         Indicated for an incoming transfer protected by a protection code.
         Not available for on-demand transfers.
        """
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        payload = {"operation_id": operation_id, "protection_code": protection_code}
        response = await self._request_service.api_request(
            "POST",
            YooMoneyMethods.ACCEPT_INCOMING_TRANSFER,
            self._router,
            headers=headers,
            data=payload,
        )
        return IncomingTransaction.parse_obj(response)

    async def reject_transaction(self, operation_id: str) -> Dict[str, str]:
        """
        Cancellation of incoming transfers protected by a protection code if you transferred
         in the send method the protect parameter,
        and transfers on demand. \n
        If the transfer is canceled, it is returned to the sender. \n
        Required token rights: incoming-transfers
        Docs:
        https://yoomoney.ru/docs/wallet/process-payments/incoming-transfer-reject

        :param operation_id: Operation identifier, parameter value
         operation_id of history () method response.
        """
        headers = self._auth_token(
            retrieve_base_headers_for_yoomoney(**self._router.config.content_and_auth)
        )
        return await self._request_service.api_request(
            "POST",
            YooMoneyMethods.INCOMING_TRANSFER_REJECT,
            self._router,
            headers=headers,
            data={"operation_id": operation_id},
        )

    async def check_transaction(
            self,
            amount: Union[int, float],
            operation_type: str = "in",
            comment: Optional[str] = None,
            rows: int = 100,
            recipient: Optional[str] = None,
    ) -> bool:
        """
        Method for verifying a transaction.
        This method uses self.transactions (rows = rows) to receive payments.
        For a little optimization, you can decrease rows by setting it,
        however, this does not guarantee the correct result

        :param amount: payment amount
        :param operation_type: payment type
        :param recipient: recipient number
        :param rows: number of payments to be checked
        :param comment: comment by which the transaction will be verified
        :return: bool, is there such a transaction in the payment history
        """
        types = [OperationType.from_input(operation_type)]
        transactions = await self.transactions(operation_types=types, records=rows)
        tasks = [self.transaction_info(txn.id) for txn in transactions]
        for detail in await asyncio.gather(*tasks):
            # Parse amount and comment,
            # because the parameters depend on the type of transaction
            amount_, comment_ = parse_amount(operation_type, detail)
            checked = check_params(amount_, amount, detail, operation_type)
            if checked and detail.status == "success":
                if comment_ == comment and detail.sender == recipient:
                    return True
                elif isinstance(comment, str) and isinstance(recipient, str):
                    continue
                elif comment_ == comment:
                    return True

        return False



from __future__ import annotations

import contextvars
from typing import (
    Any,
    TypeVar,
    Optional,
    cast,
    Generic,
    ClassVar,
    Dict,
    Type,
    Union, )

from glQiwiApi.core.dispatcher.filters import BaseFilter
from glQiwiApi.core.dispatcher.implementation import (
    Dispatcher,
    TxnRawHandler,
    TxnFilters,
    BillFilters,
    BillRawHandler,
    ErrorRawHandler,
)


class DataMixin:
    @property
    def config_data(self) -> Dict[Any, Any]:  # pragma: no cover
        data = getattr(self, "_data", None)
        if data is None:
            data = {}
            setattr(self, "_data", data)
        return cast(Dict[Any, Any], data)

    def __getitem__(self, item: Any) -> Any:
        return self.config_data[item]  # pragma: no cover

    def __setitem__(self, key: Any, value: Any) -> None:
        self.config_data[key] = value  # pragma: no cover

    def __delitem__(self, key: Any) -> None:
        del self.config_data[key]  # pragma: no cover

    def __contains__(self, key: Any) -> bool:
        return key in self.config_data  # pragma: no cover

    def get(self, key, default=None):  # type: ignore
        return self.config_data.get(key, default)  # pragma: no cover


ContextInstance = TypeVar("ContextInstance")


class ContextInstanceMixin(Generic[ContextInstance]):
    __context_instance: ClassVar[contextvars.ContextVar[ContextInstance]]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__()
        cls.__context_instance = contextvars.ContextVar(f"instance_{cls.__name__}")

    @classmethod  # noqa: F811
    def get_current(  # noqa: F811
            cls, no_error: bool = True
    ) -> Optional[ContextInstance]:  # pragma: no cover  # noqa: F811
        """Get current instance from context"""
        # on mypy 0.770 I catch that contextvars.ContextVar always contextvars.ContextVar[Any]
        cls.__context_instance = cast(
            contextvars.ContextVar[ContextInstance], cls.__context_instance
        )

        try:
            current: Optional[ContextInstance] = cls.__context_instance.get()
        except LookupError:
            if no_error:
                current = None
            else:
                raise

        return current

    @classmethod
    def set_current(cls, value: ContextInstance) -> contextvars.Token[ContextInstance]:
        if not isinstance(value, cls):
            raise TypeError(  # pragma: no cover
                f"Value should be instance of {cls.__name__!r} not {type(value).__name__!r}"
            )
        return cls.__context_instance.set(value)

    @classmethod
    def reset_current(cls, token: contextvars.Token[ContextInstance]) -> None:
        cls.__context_instance.reset(token)


class HandlerCollectionMixin:
    _dispatcher: Optional[Dispatcher]

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__()
        cls._dispatcher = None

    @property
    def dispatcher(self) -> Dispatcher:
        if not self._dispatcher:
            self._dispatcher = Dispatcher()
        return self._dispatcher

    @dispatcher.setter
    def dispatcher(self, other: Dispatcher) -> None:
        if not isinstance(other, Dispatcher):
            raise TypeError(f"Expected `Dispatcher`, got wrong type {type(other)}")
        self._dispatcher = other

    @property
    def transaction_handler(self):  # type: ignore
        """
        Handler manager for default QIWI transactions,
        you can pass on lambda filter, if you want,
        but it must to return a boolean
        """
        return self.dispatcher.transaction_handler

    @property
    def error_handler(self):  # type: ignore
        return self.dispatcher.error_handler

    @property
    def bill_handler(self):  # type: ignore
        """
        Handler manager for P2P bills,
        you can pass on lambda filter, if you want
        But it must to return a boolean
        """
        return self.dispatcher.bill_handler

    def register_transaction_handler(
            self, event_handler: TxnRawHandler, *filters: TxnFilters
    ) -> None:
        return self.dispatcher.register_transaction_handler(event_handler, *filters)

    def register_bill_handler(
            self, event_handler: BillRawHandler, *filters: BillFilters
    ) -> None:
        return self.dispatcher.register_bill_handler(event_handler, *filters)

    def register_error_handler(
            self,
            event_handler: ErrorRawHandler,
            exception: Optional[Union[Type[Exception], Exception]] = None,
            *filters: BaseFilter[Exception],
    ) -> None:
        return self.dispatcher.register_error_handler(
            event_handler, exception, *filters
        )



from __future__ import annotations

import abc
import copy
from types import TracebackType
from typing import Union, Optional, Type, Any, Dict, cast, TypeVar, TYPE_CHECKING

from glQiwiApi.core.constants import NO_CACHING
from glQiwiApi.core.storage import UnrealizedCacheInvalidationStrategy

if TYPE_CHECKING:
    from glQiwiApi.core.request_service import RequestService

_T = TypeVar("_T", bound="Wrapper")


class Wrapper(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def get_request_service(self) -> RequestService:
        pass

    def enable_caching(self, *, cache_time_in_seconds: Union[int, float]) -> Wrapper:
        request_service = self.get_request_service()
        invalidate_strategy = request_service._cache._invalidate_strategy  # noqa
        old_cache_time: Optional[Union[int, float]] = getattr(
            invalidate_strategy, "_cache_time", None
        )
        if old_cache_time == NO_CACHING:
            setattr(invalidate_strategy, "_cache_time", cache_time_in_seconds)
        return self

    def disable_caching(self) -> Wrapper:
        request_service = self.get_request_service()
        request_service._cache._invalidate_strategy = UnrealizedCacheInvalidationStrategy()
        return self

    async def __aenter__(self):  # type: ignore
        await self.get_request_service().warmup()
        return self

    async def __aexit__(
            self,
            exc_type: Optional[Type[BaseException]],
            exc_value: Optional[BaseException],
            traceback: Optional[TracebackType],
    ) -> None:
        pass
        await self.close()

    async def close(self) -> None:
        await self.get_request_service().shutdown()

    def _get(self, item: Any) -> Any:  # pragma: no cover
        try:
            return super().__getattribute__(item)
        except AttributeError:
            return None

    def __deepcopy__(self, memo: Dict[Any, Any]) -> Wrapper:  # pragma: no cover
        cls = self.__class__
        kw = {"__copy_signal__": True}
        result = cls.__new__(cls, **kw)  # type: ignore  # pragma: no cover
        memo[id(self)] = result
        instance_dictionary = {
            slot: self._get(slot)
            for slot in ["_request_service", "dispatcher"]
            if self._get(slot) is not None
        }
        for k, value in instance_dictionary.items():
            if k == "_request_service":
                value._holder = None
            elif k == "dispatcher":
                value._loop = None
            setattr(result, k, copy.deepcopy(value, memo))  # NOQA
        return cast(Wrapper, result)



from __future__ import annotations

import typing as t
from ssl import SSLContext

from aiohttp import web

from glQiwiApi.plugins.telegram.base import TelegramPlugin

if t.TYPE_CHECKING:
    from glQiwiApi.utils.compat import Dispatcher

# Some aliases =)
ListOfRoutes = t.List[web.ResourceRoute]
SubApps = t.List[t.Tuple[str, web.Application, ListOfRoutes]]


def _init_sub_apps_handlers(app: web.Application, routes: ListOfRoutes) -> None:
    """
    Initialize sub application handlers

    :param app:
    :param routes: list of AbstractRoute subclasses
    """
    for route in routes:  # pragma: no cover
        app.router.add_route(
            handler=route.handler,
            method=route.method,
            name=route.name,
            path=route.url_for().path,
        )


class TelegramWebhookPlugin(TelegramPlugin):
    """
    Managing loading webhooks of aiogram together with QiwiWrapper

    """

    execution_path: str = "/bot/{token}"
    """You can override this parameter to change the default route path"""

    prefix: str = "/tg/webhooks"
    """ You can override the prefix for the application """

    def __init__(
            self,
            dispatcher: Dispatcher,
            ssl_certificate: SSLContext,
            webhook_domain: str,
            route_name: str = "webhook_handler",
            sub_apps: t.Optional[SubApps] = None
    ) -> None:
        """

        :param dispatcher: instance of aiogram class Dispatcher
        :param ssl_certificate: self-signed or default ssl certificate
        :param webhook_domain: your domain such like https://example.com
        :param route_name: name of aiogram webhook router
        :param sub_apps: list of tuples(prefix as string, web.Application)
        """
        from aiogram.dispatcher.webhook import WebhookRequestHandler

        super(TelegramWebhookPlugin, self).__init__(dispatcher)
        self._app: web.Application = web.Application()
        self._app.router.add_route(
            "*", self.execution_path, WebhookRequestHandler, name=route_name
        )
        self._app["BOT_DISPATCHER"] = self.dispatcher
        self.sub_apps: SubApps = sub_apps or []
        self._ssl_context = ssl_certificate
        self._webhook_domain = webhook_domain

    async def incline(self, ctx: t.Dict[t.Any, t.Any]) -> web.Application:
        """
        A method that connects the main application, and the proxy in the form of a telegram

        :param ctx: keyword arguments, which contains application and host
        """
        main_app = t.cast(web.Application, ctx.get("app"))
        main_app.add_subapp(self.prefix, self._app)
        for prefix, sub_app, handlers in self.sub_apps:
            sub_app["bot"] = self.bot
            sub_app["dp"] = self.dispatcher
            _init_sub_apps_handlers(sub_app, handlers)
            main_app.add_subapp(prefix, sub_app)
        await self.configure_webhook(ctx)

        return self._app

    async def configure_webhook(self, ctx: t.Dict[t.Any, t.Any]) -> None:
        """
        You can override this method to correctly setup webhooks with aiogram
        API method `set_webhook` like this: self.dispatcher.bot.set_webhook()

        """
        full_url = self._webhook_domain + self.prefix
        if isinstance(self.execution_path, str):
            full_url += self.execution_path
        await self.dispatcher.bot.set_webhook(full_url, certificate=self._ssl_context, **ctx)

    async def shutdown(self) -> None:
        await self.dispatcher.storage.close()
        await self.dispatcher.storage.wait_closed()
        await self.dispatcher.bot.session.close()



from __future__ import annotations

import json
from typing import Optional, Union, Dict, Any, TYPE_CHECKING

from aiohttp import RequestInfo
from pydantic import BaseModel

if TYPE_CHECKING:
    from glQiwiApi.types.errors import QiwiErrorAnswer


class CantParseUrl(Exception):
    pass


class NetworkError(Exception):
    pass


class InvalidPayload(TypeError):
    pass


class NoUpdatesToExecute(Exception):
    pass


class InvalidCachePayload(Exception):
    pass


class RequestInfoModel(BaseModel):
    method: str
    url: str
    real_url: str


class ExceptionTraceback(BaseModel):
    status_code: int
    msg: Optional[str] = None
    additional_info: Optional[str] = None
    request_info: Optional[RequestInfoModel] = None


class ChequeIsNotAvailable(Exception):
    def __init__(self, err_model: QiwiErrorAnswer):
        self.error_model = err_model


class APIError(Exception):
    def __init__(
            self,
            message: Optional[str],
            status_code: Union[str, int],
            additional_info: Optional[str] = None,
            traceback_info: Optional[Union[RequestInfo, str, bytes, Dict[Any, Any]]] = None,
    ) -> None:
        super(APIError, self).__init__()
        self.message = message
        self.status_code = status_code
        self.additional_info = additional_info
        self.traceback_info = traceback_info

    def __str__(self) -> str:
        resp = "code={sc} doc(for specific cases may be deceiving)={msg}, additional_info={info}" ""
        return resp.format(
            sc=self.status_code, msg=self.message, info=self.additional_info
        )

    def to_model(self) -> ExceptionTraceback:
        """Convert exception to :class:`ExceptionTraceback`"""
        if not isinstance(self.traceback_info, RequestInfo):
            raise TypeError(
                "Cannot convert exception to `ExceptionTraceback`, because "
                "this method require `RequestInfo` object"
            )
        return ExceptionTraceback(
            status_code=self.status_code,
            msg=self.message,
            additional_info=self.additional_info,
            request_info=RequestInfoModel(
                method=self.traceback_info.method,
                url=self.traceback_info.url.__str__(),
                real_url=self.traceback_info.real_url.__str__(),
            ),
        )

    def _make_dict(self) -> Dict[str, Any]:
        return {
            "code": self.status_code,
            "msg": self.message,
            "additional_info": self.additional_info,
            "traceback_info": self.traceback_info,
        }

    def json(self, indent: int = 4, **dump_kw: Any) -> str:
        """
        Method, that makes json format from traceback

        :param indent:
        :param dump_kw:
        """
        if isinstance(self.traceback_info, RequestInfo):
            info = self.to_model().dict(exclude_none=True)
        else:
            info = self._make_dict()
        return json.dumps(info, indent=indent, ensure_ascii=False, **dump_kw)


class BadCallback(Exception):
    ...


__all__ = (
    "InvalidPayload",
    "CantParseUrl",
    "APIError",
    "NoUpdatesToExecute",
    "NetworkError",
    "InvalidCachePayload",
    "BadCallback",
    "ChequeIsNotAvailable"
)



from __future__ import annotations

import logging
import uuid
from typing import Dict, Optional, Any, Union, cast

from aiohttp import ClientTimeout, ServerDisconnectedError, \
    ClientProxyConnectionError, ClientConnectionError, RequestInfo, ClientSession
from aiohttp.typedefs import LooseCookies

from glQiwiApi.core.abc.router import AbstractRouter
from glQiwiApi.core.constants import NO_CACHING
from glQiwiApi.core.session.holder import AbstractSessionHolder, AiohttpSessionHolder
from glQiwiApi.core.storage import InMemoryCacheStorage, APIResponsesCacheInvalidationStrategy, \
    CachedAPIRequest, Payload
from glQiwiApi.utils.exceptions import APIError
from glQiwiApi.utils.payload import make_payload, get_decoded_result

logger = logging.getLogger("glQiwiApi.RequestService")


class EmptyMessages(Dict[Any, Any]):
    EMPTY_STRING = ""

    def __init__(self) -> None:
        super().__init__({})

    def __getattr__(self, item: Any) -> Any:
        return self.EMPTY_STRING


class RequestService:
    """
    Deal with :class:`Storage`,
    caching queries and managing stable work of sending requests

    """

    def __init__(
            self,
            error_messages: Optional[Dict[int, str]] = None,
            cache_time: Union[float, int] = NO_CACHING,
            session_holder: Optional[AbstractSessionHolder[Any]] = None,
    ) -> None:
        if session_holder is None:
            session_holder = AiohttpSessionHolder(
                timeout=ClientTimeout(total=5, connect=None, sock_connect=5, sock_read=None)
            )
        self.error_messages = error_messages or EmptyMessages()
        self._cache = InMemoryCacheStorage(
            invalidate_strategy=APIResponsesCacheInvalidationStrategy(cache_time=cache_time),
        )
        self._session_holder: AbstractSessionHolder[Any] = session_holder

    async def api_request(
            self,
            http_method: str,
            api_method: str,
            router: AbstractRouter,
            set_timeout: bool = True,
            cookies: Optional[LooseCookies] = None,
            json: Optional[Any] = None,
            data: Optional[Dict[Any, Any]] = None,
            headers: Optional[Dict[Any, Any]] = None,
            params: Optional[Dict[Any, Any]] = None,
            **kwargs: Any
    ) -> Dict[Any, Any]:
        url = router.build_url(api_method, **kwargs)
        return await self.raw_request(
            url,
            http_method,
            set_timeout=set_timeout,
            headers=headers,
            json=json,
            data=data,
            params=params,
            cookies=cookies,
        )

    async def raw_request(
            self,
            url: str,
            method: str,
            set_timeout: bool = True,
            cookies: Optional[LooseCookies] = None,
            json: Optional[Any] = None,
            data: Optional[Any] = None,
            headers: Optional[Any] = None,
            params: Optional[Any] = None,
            **kwargs: Any
    ) -> Dict[Any, Any]:
        request_args = {k: v for k, v in locals().items() if not isinstance(v, type(self))}
        if await self._storage_has_similar_cached_response(**request_args):
            return (await self._cache.retrieve(url)).response  # type: ignore
        session = await self._session_holder.get_session()
        _request_id = uuid.uuid4().hex
        try:
            logger.debug(
                "Send request %s to %s body = %s params = %s headers = %s json = %s",
                _request_id, url, data, params, headers, json
            )
            http_response = await session.request(
                method=method,
                url=url,
                data=data,
                headers=headers,
                json=json if isinstance(json, dict) else None,
                cookies=cookies,
                params=params,
                **kwargs,
            )
            decoded_response = get_decoded_result(
                self.error_messages,
                http_response.status,
                http_response.request_info,
                await http_response.text(),
            )
        except (
                ClientProxyConnectionError,
                ServerDisconnectedError,
                ClientConnectionError,
        ):
            raise self.api_exception(status_code=500)
        else:
            logger.debug(
                "Request %s is successful status = %d decoded response = %s",
                _request_id,
                http_response.status,
                decoded_response
            )
            self._cache_result(decoded_response, **request_args)
            return decoded_response

    async def text_content(
            self,
            url: str,
            method: str,
            cookies: Optional[LooseCookies] = None,
            json: Optional[Any] = None,
            data: Optional[Any] = None,
            headers: Optional[Any] = None,
            params: Optional[Any] = None,
    ) -> str:
        logger.debug("Get text content from %s method = %s json = %s body = %d headers = %s params = %s ",
                     url, method, json, data, headers, params)
        prepared_payload = make_payload(**locals())
        session: ClientSession = await self._session_holder.get_session()
        resp = await session.request(**prepared_payload)
        return cast(str, resp)

    async def retrieve_bytes(self, url: str, method: str, **kwargs: Any) -> bytes:
        session = await self._session_holder.get_session()
        resp = await session.request(method, url, **kwargs)
        return cast(bytes, await resp.read())

    async def _storage_has_similar_cached_response(self, **request_args: Any) -> bool:
        return await self._cache.contains_similar(Payload(**request_args))

    def _cache_result(self, response: Any, method: str, **kwargs: Any) -> None:
        logger.debug("Cache response %s kwargs = %s", response, kwargs)
        self._cache.update(
            **{
                kwargs["url"]: CachedAPIRequest(
                    payload=Payload(**kwargs),
                    response=response,
                    method=method
                )
            }
        )

    def api_exception(
            self,
            status_code: int,
            traceback_info: Optional[Union[RequestInfo, Dict[Any, Any], str, bytes]] = None,
            message: Optional[str] = None,
    ) -> APIError:
        """Raise :class:`APIError` exception with pretty explanation"""
        from glQiwiApi import __version__

        if not isinstance(message, str) and isinstance(self.error_messages, dict):
            message = self.error_messages.get(status_code, "Unknown")
        return APIError(
            message,
            status_code,
            additional_info=f"{__version__} API version",
            traceback_info=traceback_info,
        )

    def reset_cache(self) -> None:
        self._cache.clear()

    async def warmup(self) -> Any:
        return await self._session_holder.get_session()

    async def shutdown(self) -> None:
        await self._session_holder.close()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glQiwiApi',
 'glQiwiApi.contrib',
 'glQiwiApi.core',
 'glQiwiApi.core.abc',
 'glQiwiApi.core.dispatcher',
 'glQiwiApi.core.dispatcher.class_based',
 'glQiwiApi.core.dispatcher.webhooks',
 'glQiwiApi.core.dispatcher.webhooks.views',
 'glQiwiApi.core.session',
 'glQiwiApi.core.synchronous',
 'glQiwiApi.ext',
 'glQiwiApi.plugins',
 'glQiwiApi.plugins.telegram',
 'glQiwiApi.qiwi',
 'glQiwiApi.types',
 'glQiwiApi.types.arbitrary',
 'glQiwiApi.types.qiwi',
 'glQiwiApi.types.yoomoney',
 'glQiwiApi.utils',
 'glQiwiApi.yoo_money']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.0,<4.0.0',
 'pydantic==1.8.2',
 'pytz==2021.3',
 'wheel>=0.36.2,<0.37.0']

extras_require = \
{':python_version <= "3.7"': ['typing_extensions>=3.10.0.2,<4.0.0.0']}

setup_kwargs = {
    'name': 'glqiwiapi',
    'version': '1.1.4',
    'description': 'The ultrarapid and multifunctional wrapper over QIWI and YooMoney',
    'long_description': '<h2 align="center">\n<img src="https://github.com/GLEF1X/glQiwiApi/blob/master/docs/static/logo.png" width="200"></img>\n\n\n[![PyPI version](https://img.shields.io/pypi/v/glQiwiApi.svg)](https://pypi.org/project/glQiwiApi/) [![Code Quality Score](https://www.code-inspector.com/project/20780/score/svg)](https://frontend.code-inspector.com/public/project/20780/glQiwiApi/dashboard) ![Code Grade](https://www.code-inspector.com/project/20780/status/svg) ![Downloads](https://img.shields.io/pypi/dm/glQiwiApi) ![docs](https://readthedocs.org/projects/pip/badge/?version=latest)\n[![Language grade: Python](https://img.shields.io/lgtm/grade/python/g/GLEF1X/glQiwiApi.svg?logo=lgtm&logoWidth=18)](https://lgtm.com/projects/g/GLEF1X/glQiwiApi/context:python) [![CodeFactor](https://www.codefactor.io/repository/github/glef1x/glqiwiapi/badge)](https://www.codefactor.io/repository/github/glef1x/glqiwiapi)\n![codecov](https://codecov.io/gh/GLEF1X/glQiwiApi/branch/dev-1.x/graph/badge.svg?token=OD538HKV15)\n![CI](https://github.com/GLEF1X/glQiwiApi/actions/workflows/tests.yml/badge.svg) ![mypy](https://img.shields.io/badge/%20type_checker-mypy-%231674b1?style=flat) [![Downloads](https://pepy.tech/badge/glqiwiapi/month)](https://pepy.tech/project/glqiwiapi) [![Downloads](https://pepy.tech/badge/glqiwiapi)](https://pepy.tech/project/glqiwiapi)\n\n<img src="https://github.com/GLEF1X/glQiwiApi/blob/master/demo.gif"/>\n</h2>\n\n## 🌎Official api resources:\n\n* 🎓 __Documentation: [here](https://glqiwiapi.readthedocs.io/en/latest/)__\n* 🖱️ __Developer\n  contacts: [![Dev-Telegram](https://img.shields.io/badge/Telegram-blue.svg?style=flat-square&logo=telegram)](https://t.me/GLEF1X)__\n\n## 🐦Dependencies\n\n| Library |                       Description                       |\n|:-------:|:-------------------------------------------------------:|\n|aiohttp  | Asynchronous HTTP Client/Server for asyncio and Python. |\n|aiofiles |            saving receipts in pdf(Optional)             |\n|pydantic |    Json data validator. Very fast instead of custom     |\n',
    'author': 'Gleb Garanin',
    'author_email': 'glebgar567@gmail.com',
    'maintainer': 'GLEF1X',
    'maintainer_email': 'glebgar567@gmail.com',
    'url': 'https://github.com/GLEF1X/glQiwiApi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import functools as ft
import inspect
import time
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from glQiwiApi import QiwiWrapper


class measure_time(object):  # NOQA
    def __init__(self, logger=None):
        self._logger = logger

    def __call__(self, func):
        if inspect.iscoroutinefunction(func) or inspect.iscoroutine(func):

            @ft.wraps(func)
            async def wrapper(*args, **kwargs):
                start_time = time.monotonic()
                result = await func(*args, **kwargs)
                execute_time = time.monotonic() - start_time
                msg = "Function `%s` executed for %s secs"
                self._log(msg, func.__name__, execute_time)
                return result

        else:

            @ft.wraps(func)
            def wrapper(*args, **kwargs):
                start_time = time.monotonic()
                result = func(*args, **kwargs)
                execute_time = time.monotonic() - start_time
                msg = "Function `%s` executed for %s secs"
                self._log(msg, func.__name__, execute_time)
                return result

        return wrapper

    def _log(self, msg, *args):
        if self._logger is not None:
            self._logger.info(msg, *args)
        else:
            print(msg % args)


class allow_response_code:  # NOQA
    def __init__(self, status_code) -> None:
        self.status_code = status_code

    def __call__(self, func):
        status_code = self.status_code

        @ft.wraps(func)
        async def wrapper(*args, **kwargs):
            from glQiwiApi import APIError

            try:
                await func(*args, **kwargs)
            except APIError as error:
                if error.status_code == str(status_code):
                    info = error.traceback_info
                    return {"success": True} if not info else info
                return {"success": False}

        return wrapper


class override_error_messages:  # NOQA
    def __init__(self, status_codes) -> None:
        self.status_codes = status_codes

    def __call__(self, func):
        status_codes = self.status_codes

        @ft.wraps(func)
        async def wrapper(*args, **kwargs):
            from glQiwiApi import APIError

            try:
                return await func(*args, **kwargs)
            except APIError as ex:
                if int(ex.status_code) in status_codes.keys():
                    error = status_codes.get(int(ex.status_code))
                    ex = APIError(
                        message=error.get("message"),
                        status_code=ex.status_code,
                        traceback_info=error.get("json_info"),
                        additional_info=ex.additional_info,
                    )
                raise ex from None

        return wrapper


class require:

    def __init__(self, *params):
        self._required_attrs = params

    def __call__(self, func):
        @ft.wraps(func)
        async def wrapper(c: QiwiWrapper, *args, **kwargs):
            self.check_is_object_contains_required_attrs(c, func)
            return await func(c, *args, **kwargs)

        return wrapper

    def check_is_object_contains_required_attrs(self, c, func):
        from glQiwiApi import InvalidPayload

        for required_attr_name in self._required_attrs:
            if getattr(c, required_attr_name, None) is None:
                raise InvalidPayload(
                    f"Method {func.__name__} requires {required_attr_name} not to be empty"
                )



from __future__ import annotations

from datetime import datetime
from typing import Optional, cast

import pytz

from glQiwiApi.utils.compat import Final

DEFAULT_QIWI_TIMEZONE: Final[str] = "Europe/Moscow"


def datetime_to_utc(obj: datetime) -> str:
    iso_format_date = pytz.utc.localize(obj).replace(tzinfo=None).isoformat(" ")
    return iso_format_date.replace(" ", "T") + "Z"


def datetime_to_iso8601_with_moscow_timezone(obj: Optional[datetime]) -> str:
    """
    Converts a date to a standard format for API's

    :param obj: datetime object to parse to string
    :return: string - parsed date
    """
    if not isinstance(obj, datetime):
        return ""  # pragma: no cover
    return obj.astimezone(pytz.timezone(DEFAULT_QIWI_TIMEZONE)).isoformat(timespec='seconds')



from __future__ import annotations

import ipaddress

from glQiwiApi.core.dispatcher.webhooks.config import ALLOWED_IPS


def check_ip(ip_address: str) -> bool:
    """
    Check if ip is allowed to request us

    :param ip_address: IP-address
    :return: address is allowed
    """
    address = ipaddress.IPv4Address(ip_address)
    unpacked = [ip_address for pool in ALLOWED_IPS for ip_address in pool]
    return address in unpacked



import base64
import hashlib
import hmac
import warnings
from datetime import datetime
from typing import Optional, Union, Dict, Any

from pydantic import Field, Extra

from glQiwiApi.types.amount import PlainAmount, HashableOptionalSum
from glQiwiApi.types.base import Base, HashableBase
from glQiwiApi.types.exceptions import WebhookSignatureUnverified


class Customer(HashableBase):
    """Object: Customer"""

    phone: Optional[str] = None
    email: Optional[str] = None
    account: Optional[str] = None


class BillStatus(HashableBase):
    """Object: BillStatus"""

    value: str
    changed_datetime: Optional[datetime] = Field(None, alias="changedDateTime")


class CustomFields(HashableBase):
    """Object: CustomFields"""

    pay_sources_filter: Optional[str] = Field(None, alias="paySourcesFilter")
    theme_code: Optional[str] = Field(None, alias="themeCode")


class BillError(HashableBase):
    """Object: BillError"""

    service_name: str = Field(..., alias="serviceName")
    error_code: str = Field(..., alias="errorCode")
    description: str
    user_message: str = Field(..., alias="userMessage")
    datetime: str = Field(..., alias="dateTime")
    trace_id: str = Field(..., alias="traceId")


class Bill(HashableBase):
    """Object: Bill"""

    site_id: str = Field(..., alias="siteId")
    bill_id: str = Field(..., alias="billId")
    amount: HashableOptionalSum
    status: BillStatus
    creation_date_time: Optional[datetime] = Field(None, alias="creationDateTime")
    expiration_date_time: Optional[datetime] = Field(None, alias="expirationDateTime")
    pay_url: Optional[str] = Field(None, alias="payUrl")
    custom_fields: Optional[CustomFields] = Field(None, alias="customFields")
    customer: Optional[Customer] = None
    workaround_url: Optional[str] = None

    class Config:
        extra = Extra.allow
        allow_mutation = True

    async def check(self) -> bool:
        """Checking p2p payment status"""
        return (await self.client.check_p2p_bill_status(bill_id=self.bill_id)) == "PAID"

    async def reject(self) -> None:
        await self.client.reject_p2p_bill(bill_id=self.bill_id)


class RefundBill(Base):
    """object: RefundBill"""

    amount: PlainAmount
    datetime: datetime
    refund_id: str = Field(..., alias="refundId")
    status: str

    def as_str(self) -> str:
        return f"№{self.refund_id} {self.status} {self.amount} {self.datetime}"

    def get_value(self) -> Union[float, int]:
        return self.amount.value


class Notification(HashableBase):
    """Object: Notification"""

    version: str = Field(..., alias="version")
    bill: Bill = Field(..., alias="bill")

    def __repr__(self) -> str:
        return f"#{self.bill.bill_id} {self.bill.amount} {self.bill.status} "

    def verify_signature(self, sha256_signature: str, webhook_base64_key: str) -> None:
        webhook_key = base64.b64decode(bytes(webhook_base64_key, "utf-8"))
        bill = self.bill

        invoice_params = f"{bill.amount.currency}|{bill.amount.value}|{bill.bill_id}|{bill.site_id}|{bill.status.value}"
        generated_signature = hmac.new(
            webhook_key, invoice_params.encode("utf-8"), hashlib.sha256
        ).hexdigest()

        if generated_signature != sha256_signature:
            raise WebhookSignatureUnverified()


class P2PKeys(Base):
    public_key: str = Field(..., alias="PublicKey")
    secret_key: str = Field(..., alias="SecretKey")


class InvoiceStatus(Base):
    invoice_status: str
    is_sms_confirm: bool
    pay_results: Dict[Any, Any] = Field(..., alias="WALLET_ACCEPT_PAY_RESULT")


__all__ = (
    "Bill",
    "BillError",
    "RefundBill",
    "Notification",
    "P2PKeys",
    "InvoiceStatus",
    "BillStatus",
)
